rm(list=ls())
library(Rcpp)
library(RcppArmadillo)
library(fields)
sourceCpp("MCMCfunction.cpp")
sourceCpp("MCMCsampler.cpp")

#Function that computes the CP consensus
CPconsensus<-function(mcmcOutput,labs){
  
  consensus <- vector(length=n)
  if(is.character(labs)){ items<-labs}else{
    paste("0",1:n,sep="")
  }
  names(consensus)<-items
  nmceff<-dim(mcmcOutput)[1]
  lim=(nmceff/10):nmceff
  nout <- length(lim)
  rem.items <- items
  dimnames(mcmcOutput)[[2]]<-items
  
  for(i in 1:(n-1)){
    chi<-colSums(mcmcOutput[lim,rem.items]<=i)/nout
    ind.item <- which.max(chi)
    which.item <- rem.items[ind.item]
    consensus[which.item] <- i
    rem.items <- rem.items[-ind.item]
  }
  consensus[rem.items] <- n
  
  return(consensus)
} 
#Function that computes matrix of posterior marginal probabilities for each item
heatMat<-function(mcmcOutput,burnin,t_rank){
  n<-dim(mcmcOutput)[2]
  lim1<-burnin:dim(mcmcOutput)[1]
  f<-matrix(nrow=n, ncol=n)
  colnames(mcmcOutput)<-NULL
  for(j in 1:n){
    ind <- which(t_rank==j)
    f[j, ] <- sapply(1:n, function(x) sum(mcmcOutput[lim1,ind] == x))
  }
  f <- f/length(lim1)
  return(f)
}
#Function that created the heatplot
heatPlot<-function(mat,t_rank){
  
  if(is.character(names(t_rank))){
    items<-names(sort(t_rank))}else{
      items<-paste("0",order(t_rank),sep="")
    }
  
  par(mfrow=c(1,1))
  par(mar=c(2,2,2,6))
  image(mat,col=tim.colors(64*10),zlim=range(mat),axes=F,cex.lab=2)
  
  mtext(text=items, side=1, line=0.3, 
        at=seq(0,1,1/(n-1)), las=1, cex=1)
  mtext(text=c(1:n), side=2, line=0.3, 
        at=seq(0,1,1/(n-1)), las=2, cex=1)
  par(mar=c(2,2,2,2))
  image.plot(mat, legend.only=T,horizontal = F)
}



###  Full ranks (without clusters) ###  

#A# Sample N full rankings from the Mallows model
#   with a given distance (metric) 
#   Algorithm 5 of the paper

# choice of the model parameters
N<-500  # number of assessors
n<-50   # number of items
alfa<-10
latrank<-sample(n,n)
labs<-paste("A",1:n,sep="")
names(latrank)<-labs
metric="footrule"
# sampling
pars <- list("nmc"=N*10^3+1, "thin"=10^3, "burnin"=10^5, "L"=n/2-1, 
               "dist"=metric, "K"=N)
R<-MCMCSampler(latrank,alfa,pars)$R 

#B# Estimate the model parameters via MCMC 
#   Algorithm 1 of the paper

# choose MCMC parameters
nmc <- 1e5
thin <- 1e1
alphaJump <- 1e1
lambda <- 0.1
L <- n/5
sdAlpha <- 0.4

# partition function
if(metric=="Kendall"){# closed form available
  fitvec <- 0
  seq2 <- 0
}else if(metric=="footrule"){
  if(n <=50){# closed form available
    load("CdFootrule.Rdata")
    fitvec <- 0
  }else{# for larger n, run the IS
    seq2 <- 0
    sourceCpp("./importanceSampler.cpp")
    propDist   <- metric # metric in the Mallows model used in the pseudo-likelihood IS proposal
    targetDist <- metric # metric in the Mallows model used for computing Z
    K <- 1e6             # number of IS iterations
    alpha     <- seq(.01,10,length.out=50) # values of alpha for which the IS approximation has to be computed
    ## parameters for the IS run:
    ## Z (defaults to a vector of 0s) is a pre-computed vector of partition function values on the same alpha grid
    ## and K is the corresponding number of IS iterations used to compute it
    pars <- list("nmc"=K, "n"=n, "propDist"=propDist, "targetDist"=targetDist, "K"=0, "Z"=rep(0,length(alpha)))
    y <- importanceSampler(alpha, pars)    
    x <- alpha
    # approximate Z via a polynomial function of degree 10
    fitvec <- lm(y ~ x + I(x^2) + I(x^3) + I(x^4)+ I(x^5) +  I(x^6) + I(x^7) + I(x^8) + I(x^9) + I(x^10))$coefficients
  }
}else if (metric=="spearman"){# closed form available
  if(n <=14){
    load("CdSpearman.Rdata")
    fitvec <- 0
  }else{# for larger n, run the IS
    seq2 <- 0
    sourceCpp("./importanceSampler.cpp")
    propDist   <- metric
    targetDist <- metric
    K <- 1e6
    alpha     <- seq(.01,20,length.out=50)
    pars <- list("nmc"=K, "n"=n, "propDist"=propDist, "targetDist"=targetDist, "K"=0, "Z"=rep(0,length(alpha)))
    y <- importanceSampler(alpha, pars)    
    x <- alpha
    fitvec <- lm(y ~ x + I(x^2) + I(x^3) + I(x^4)+ I(x^5) +  I(x^6) + I(x^7) + I(x^8) + I(x^9) + I(x^10))$coefficients
  }
}

# run the MCMC
pars <- list("nmc"=nmc, "thin"=thin, "alphaJump"=alphaJump, 
             "lambda"=lambda, "L"=L, "sdAlpha"=sdAlpha, 
             "dist"=metric, "fit"=fitvec,"aug"=FALSE, 
             "Rmiss"=sample(n),"alpha0"=2, "rho0"=sample(n),
             "Cd"=seq2)
result <- MCMCfunction(R, pars)

#Acceptance
result$accRho
result$accAlpha
#Trace of Alpha
plot(result$alpha,type = "l")
#choose burnin and reduce the length of the chain
burnin<-100
nout<-dim(result$rho)[1]
lim<-burnin:nout
#estimated CP consensus
cpcons<-CPconsensus(result$rho[lim,],labs)
names(cpcons)<-labs
rbind(latrank,cpcons)

#heat plot of rho (ordered according to true consensus)
matcons<-heatMat(result$rho[lim,],1,latrank)
heatPlot(matcons,latrank)

