#include <algorithm>
#include <RcppArmadillo.h>
#include <RcppArmadilloExtensions/sample.h>
//[[Rcpp::depends(RcppArmadillo)]]

using namespace Rcpp;
using namespace arma;

List leapAndShift(rowvec rho, const int& L);
double vecdist(const rowvec& R, const rowvec& rho, const std::string& metric);

// [[Rcpp::export]]
List MCMCSampler(vec rho0, double alpha0, List pars){
  RNGScope scope;
  // Forward simulation of new observations, given rho0
  // Read in all the parameters
  int nmc = as<int>(pars["nmc"]);
  int burnin = as<int>(pars["burnin"]);
  int L = as<int>(pars["L"]);
  std::string metric = as<std::string>(pars["dist"]);
  int K = as<int>(pars["K"]); // number of permutations to generate
  int thin = as<int>(pars["thin"]); // How many MCMC steps between each time we collect

  Function sample("sample");

  int n = rho0.n_elem; // number of items
  double accR = 0, ratio, u, probForw, probBack;
  mat R(K,n);
  vec Z(K);
  R.zeros();
  rowvec Rprop = trans(rho0);
  rowvec Rnow = trans(rho0);
  List leapShiftValue;
  uvec LSindices;
  
  for(int t=1; t<burnin; ++t){
    // Propose a new observation
    leapShiftValue = leapAndShift(Rnow, L);
    Rprop = as<rowvec>(leapShiftValue["prop"]);
    LSindices = as<uvec>(leapShiftValue["indices"]);
    probForw = leapShiftValue["probF"];
    probBack = leapShiftValue["probB"];
    ratio = -alpha0/n*vecdist(Rprop,trans(rho0),metric)+log(probBack)+alpha0/n*vecdist(Rnow,trans(rho0),metric)-log(probForw);
    u = log(as_scalar(randu(1,1)));
    if(ratio > u){
      Rnow = Rprop;
      accR++;
    }
  }
  std::cout << "Acceptance ratio in burn-in period = " << accR/(burnin-1) << std::endl;
  accR = 0;
  int i=0; // Number of samples taken
  for(int t=1; t<nmc; t++){
    leapShiftValue = leapAndShift(Rnow, L);
    Rprop = as<rowvec>(leapShiftValue["prop"]);
    LSindices = as<uvec>(leapShiftValue["indices"]);
    probForw = leapShiftValue["probF"];
    probBack = leapShiftValue["probB"];
    ratio = -alpha0/n*vecdist(Rprop,trans(rho0),metric)+log(probBack)+alpha0/n*vecdist(Rnow,trans(rho0),metric)-log(probForw);
    u = log(as_scalar(randu(1,1)));
    if(ratio > u){
      Rnow = Rprop;
      accR++;
    }
    if(t % thin == 0){
      R.row(i) = Rnow;
      Z(i) = vecdist(Rnow, trans(rho0), metric);
      i++;
    }
    if(i == K) break; // Then we have taken enough samples
  }

  std::cout << "Acceptance ratio = " << accR/(nmc-1) << std::endl;
  return List::create(Named("R")=R,
  	      Named("Z")=Z);

}

double vecdist(const rowvec& R, const rowvec& rho, const std::string& metric){
  
  int n = rho.n_elem;
  double totDist=0;
  if(metric == "footrule"){
    totDist = norm(R - rho, 1);
  } else if(metric == "kendall"){
    
    for(int u=0; u<n; u++){
      for(int t=0; t<u; t++){
	if((R(t) > R(u)) & (rho(t) < rho(u) ) || ((R(t) < R(u) ) & (rho(t) > rho(u)))) totDist +=1;
      }
    }
  } else if(metric == "spearman") {
    totDist = pow(norm(R - rho, 2),2.0);
  }
  return totDist;
}

List leapAndShift(rowvec rho, const int& L){
  RNGScope scope;

  rowvec prop=rho;
  IntegerVector tmp;
  vec support, indices;
  int n = rho.n_elem, u, index;
  double deltaR, probF, probB, supportNew;
  
  // Leap step: 
  // 1. sample u
  u = as_scalar(randi(1, distr_param(1,n)));
  // 2. compute the S set for sampling the new rank
  if((rho(u-1) > 1) & (rho(u-1) < n)){
    support = join_cols(linspace(std::max(1.0, rho(u-1)-L), rho(u-1)-1,std::min(rho(u-1)-1,static_cast<double>(L))), 
    linspace(rho(u-1)+1,std::min(static_cast<double>(n), rho(u-1)+L),std::min(n-rho(u-1),static_cast<double>(L))));
  } else if(rho(u-1) == 1){
    support = linspace(rho(u-1)+1,std::min(static_cast<double>(n), rho(u-1)+L),std::min(n-rho(u-1),static_cast<double>(L)));
  } else if(rho(u-1) == n){
    support = linspace(std::max(1.0, rho(u-1)-L), rho(u-1)-1,std::min(rho(u-1)-1,static_cast<double>(L)));
  }
  // 3. assign a random element of the support set, this completes the leap step
  index = as_scalar(randi(1, distr_param(0, support.n_elem-1)));
  prop(u-1) = support(index);
  
  // Compute the associated transition probabilities (BEFORE THE SHIFT STEP, WHICH IS DETERMINISTIC --> EASIER)
  if(std::abs(prop(u-1)-rho(u-1))==1){
    // in this case the transition probabilities coincide! (and in fact for L = 1 the L&S is symmetric)
    supportNew = std::min(prop(u-1)-1,static_cast<double>(L)) + std::min(n-prop(u-1),static_cast<double>(L));
    probF = 1.0/(n*support.n_elem) + 1.0/(n*supportNew);
    probB = probF;
  }else{
    // P(proposed|current)
    probF = 1.0/(n*support.n_elem);
    // P(current|proposed)
    supportNew = std::min(prop(u-1)-1,static_cast<double>(L)) + std::min(n-prop(u-1),static_cast<double>(L));
    probB = 1.0/(n*supportNew);
  }

  // Shift step: 
  deltaR = prop(u-1) - rho(u-1);
  indices = zeros(std::abs(deltaR)+1);
  indices[0] = u-1;
  if(deltaR > 0){
    for(int k=1; k<=deltaR; k++){
      index = as_scalar(find(rho == rho(u-1) + k));
      prop(index) -= 1;
      indices[k] = index;
    }
  } else if(deltaR < 0) {
    for(int k=-1; k>=deltaR; k--){
      index = as_scalar(find(rho == rho(u-1) + k));
      prop(index) += 1;
      indices[-(k)] = index;
    }
  }
  
  
  return List::create(Named("prop")=prop,
                      Named("indices")=indices,
                      Named("deltaR")=deltaR,
                      Named("probF")=probF,
                      Named("probB")=probB);
}
