rm(list=ls())
library(Rcpp)
library(RcppArmadillo)
library(fields)
require(sets)
require(relations)
library(igraph)
sourceCpp("MCMCfunction.cpp")
sourceCpp("MCMCsampler.cpp")

#Function that computes the CP consensus
CPconsensus<-function(mcmcOutput,labs){
  
  consensus <- vector(length=n)
  if(is.character(labs)){ items<-labs}else{
    paste("0",1:n,sep="")
  }
  names(consensus)<-items
  nmceff<-dim(mcmcOutput)[1]
  lim=(nmceff/10):nmceff
  nout <- length(lim)
  rem.items <- items
  dimnames(mcmcOutput)[[2]]<-items
  
  for(i in 1:(n-1)){
    chi<-colSums(mcmcOutput[lim,rem.items]<=i)/nout
    ind.item <- which.max(chi)
    which.item <- rem.items[ind.item]
    consensus[which.item] <- i
    rem.items <- rem.items[-ind.item]
  }
  consensus[rem.items] <- n
  
  return(consensus)
} 
#Function that computes matrix of posterior marginal probabilities for each item
heatMat<-function(mcmcOutput,burnin,t_rank){
  n<-dim(mcmcOutput)[2]
  lim1<-burnin:dim(mcmcOutput)[1]
  f<-matrix(nrow=n, ncol=n)
  colnames(mcmcOutput)<-NULL
  for(j in 1:n){
    ind <- which(t_rank==j)
    f[j, ] <- sapply(1:n, function(x) sum(mcmcOutput[lim1,ind] == x))
  }
  f <- f/length(lim1)
  return(f)
}
#Function that created the heatplot
heatPlot<-function(mat,t_rank){
  
  if(is.character(names(t_rank))){
    items<-names(sort(t_rank))}else{
      items<-paste("0",order(t_rank),sep="")
    }
  
  par(mfrow=c(1,1))
  par(mar=c(2,2,2,6))
  image(mat,col=tim.colors(64*10),zlim=range(mat),axes=F,cex.lab=2)
  
  mtext(text=items, side=1, line=0.3, 
        at=seq(0,1,1/(n-1)), las=1, cex=1)
  mtext(text=c(1:n), side=2, line=0.3, 
        at=seq(0,1,1/(n-1)), las=2, cex=1)
  par(mar=c(2,2,2,2))
  image.plot(mat, legend.only=T,horizontal = F)
}
#Function that returns a ranking, given an ordering 
Ord2Rank<-function(ordering){
  n<-length(ordering)
  ranking<-vector(length=n)
  names(ranking)<-paste("A",1:n,sep="")
  for(i in 1:n){
    ranking[ordering[i]]<-i
  }
  return(ranking)
}
#Function to sample the pair comparisons:
# -Inputs-
# R: Nxn matrix of full ranks 
# Mj: N-dimensional vector including the number of pairs assessed by each assessor
# -Outputs-
# a) rankings: each assessor's preferences (N-dimensional list of sets) 
# b) draws: each assessor's draws (N-dimensional list of vectors) --> this is not handled in the current version of the code! (only empty lists)
# c) constrained: each assessor's constrained items (N-dimensional list of vectors)
R2pairs<-function(R,Mj){
  ranking<-draws<-constrained<-prefs<-list()
  N<-dim(R)[1]
  n<-dim(R)[2]
  Mmax=n*(n-1)/2
  items<-t(combn(n,2))      
  
  for(j in 1:N){
    
    if(Mj[j]<2){
      ranking[[j]]<-set()
      draws[[j]] <- numeric()
      constrained[[j]] <- numeric()
      prefs[[j]] <- c()
      itemsj<-items[sample.int(Mmax,Mj[j]),]
      fav<-as.integer(which(R[j,]==min(R[j,itemsj])))
      notfav<-setdiff(itemsj,fav)
      ranking[[j]]<-set_union(ranking[[j]], set(pair(as.numeric(fav),as.numeric(notfav))))
      constrained[[j]]<-sort(unique(itemsj))
      prefs[[j]]<-c(notfav,fav)
      
    }else{
      ranking[[j]]<-set()
      draws[[j]] <- numeric()
      constrained[[j]] <- numeric()
      prefs[[j]] <- matrix(ncol=2,nrow=Mj[j])
      
      itemsj<-items[sample.int(Mmax,Mj[j]),]
      
      for(m in 1:Mj[j]){
        fav<-as.integer(which(R[j,]==min(R[j,itemsj[m,]])))
        notfav<-setdiff(itemsj[m,],fav)
        ranking[[j]]<-set_union(ranking[[j]], set(pair(as.numeric(fav),as.numeric(notfav))))
        prefs[[j]][m,]<-c(notfav,fav)
        }
      constrained[[j]]<-sort(unique(c(itemsj)))
      
    }
    
  }
  
  return(list("ranking"=ranking,"draws"=draws,"constrained"=constrained,
              "prefs"=prefs))
}
#Function to generate the transitive closure
generateTC <- function(ranking, draws, n){
  N <- length(ranking)
  tc <- list()
  for(i in 1:N){
    # randomize the draws
    if(length(draws[[i]])>0){
      for(j in 1:dim(draws[[i]])[1]){
        tmp <- sample(1:2)
        ranking[[i]] <- ranking[[i]] + set(pair(as.numeric(draws[[i]][j,tmp[1]]), as.numeric(draws[[i]][j,tmp[2]])))
      }
    }
    tmp1 <- endorelation(graph=ranking[[i]])
    tmp2 <- transitive_closure(tmp1)
    tmp3 <- relation_incidence(tmp2)
    mat <- matrix(0, nrow=n, ncol=n)
    rownames(mat) <- 1:n
    colnames(mat) <- 1:n
    mat[rownames(tmp3), colnames(tmp3)] <- tmp3
    tc[[i]] <- mat
  }
  return(tc)
}
#Function to generate the rank matrix (given the transitive closure)
generateR <- function(prefs){
  N <- length(prefs)
  n <- length(unique(as.vector(prefs[[1]])))
  Rinit <- matrix(NA, nrow=N, ncol=n) # Initial value for rank matrix
  for (j in 1:length(prefs)){
    g<-graph_from_edgelist(prefs[[j]])
    if(length(V(g))!=n)g<-add_vertices(g, setdiff(1:n,V(g)))
    Rinit[j,]<-order(topological.sort(g)[n:1])
  }
  return(Rinit)
}


###  Pair comparisons (without clusters) ###  

#A# Sample N sets of pair comparisons from the Mallows model
#   with a given distance (metric) 
#   Modification of Algorithm 5 of the paper

# choice of the model parameters
N<-50
n<-10
alfa<-4
latrank<-sample(n,n)
labs<-paste("A",1:n,sep="")
names(latrank)<-labs
metric="footrule"
# sample full ranks first
pars <- list("nmc"=N*10^3+1, "thin"=10^3, "burnin"=10^5, "L"=n/2-1, 
             "dist"=metric, "K"=N)
R<-MCMCSampler(latrank,alfa,pars)$R 
# sample the pairwise comparisons
Mmax<-n*(n-1)/2 # Maximum number of pairs
Mmin<-3         # Minimum number of pairs 
# (in order to be able to perform preference prediction we need at least 3 comparisons per assessor)
lambdaM<-20 # Mean of Poisson for random number of pairs
Mj<-rpois(N,lambdaM) # Sample number of pairs from Poisson
Mj[Mj>Mmax]<-Mmax #If I sample values larger than Mmax, replace with Mmax
Mj[Mj<Mmin]<-Mmin #If I sample values smaller than Mmin, replace with Mmin
s2<-R2pairs(R,Mj)
ranking<-s2$ranking
draws<-s2$draws
prefs<-s2$prefs
constrained<-s2$constrained
# check that all the resulting pairwise preference sets are consistent (for each assessor)
idag<-c()
g<-list()
for (j in 1:N){
  g[[j]]<-graph_from_edgelist(prefs[[j]])
  idag[j]<-is.dag(g[[j]])
}
mean(idag)
# NOTE: when using real data one may need to remove cycles (if any)

#B# Estimate the model parameters via MCMC 
#   Algorithm 3 of the paper

# choose MCMC parameters
nmc <- 1e4
thin <- 1e1
alphaJump <- 1e1
lambda <- 0.1
L <- n/5
sdAlpha <- 0.4

# partition function
if(metric=="Kendall"){# closed form available
  fitvec <- 0
  seq2 <- 0
}else if(metric=="footrule"){
  if(n <=50){# closed form available
    load("CdFootrule.Rdata")
    fitvec <- 0
  }else{# for larger n, run the IS
    seq2 <- 0
    sourceCpp("./importanceSampler.cpp")
    propDist   <- metric # metric in the Mallows model used in the pseudo-likelihood IS proposal
    targetDist <- metric # metric in the Mallows model used for computing Z
    K <- 1e6             # number of IS iterations
    alpha     <- seq(.01,10,length.out=50) # values of alpha for which the IS approximation has to be computed
    ## parameters for the IS run:
    ## Z (defaults to a vector of 0s) is a pre-computed vector of partition function values on the same alpha grid
    ## and K is the corresponding number of IS iterations used to compute it
    pars <- list("nmc"=K, "n"=n, "propDist"=propDist, "targetDist"=targetDist, "K"=0, "Z"=rep(0,length(alpha)))
    y <- importanceSampler(alpha, pars)    
    x <- alpha
    # approximate Z via a polynomial function of degree 10
    fitvec <- lm(y ~ x + I(x^2) + I(x^3) + I(x^4)+ I(x^5) +  I(x^6) + I(x^7) + I(x^8) + I(x^9) + I(x^10))$coefficients
  }
}else if (metric=="spearman"){# closed form available
  if(n <=14){
    load("CdSpearman.Rdata")
    fitvec <- 0
  }else{# for larger n, run the IS
    seq2 <- 0
    sourceCpp("./importanceSampler.cpp")
    propDist   <- metric
    targetDist <- metric
    K <- 1e6
    alpha     <- seq(.01,20,length.out=50)
    pars <- list("nmc"=K, "n"=n, "propDist"=propDist, "targetDist"=targetDist, "K"=0, "Z"=rep(0,length(alpha)))
    y <- importanceSampler(alpha, pars)    
    x <- alpha
    fitvec <- lm(y ~ x + I(x^2) + I(x^3) + I(x^4)+ I(x^5) +  I(x^6) + I(x^7) + I(x^8) + I(x^9) + I(x^10))$coefficients
  }
}

# Compute the transitive closure
tc <- generateTC(ranking, draws, n)
# Create the initial value for the MCMC
Rinit <- generateR(prefs)

# run the MCMC
pars <- list("nmc"=nmc, "thin"=thin, "alphaJump"=alphaJump, 
             "lambda"=lambda, "L"=L, "sdAlpha"=sdAlpha, 
             "dist"=metric, "fit"=fitvec,"alpha0"=2, "rho0"=sample(n),
             "Cd"=seq2, "whenTies"=nmc+1)
result <- MCMCpairwise(ranking, draws, constrained, Rinit, pars)

#Acceptance
result$accRho
result$accAlpha
result$accRaug

#Trace of Alpha
plot(result$alpha,type = "l")
#choose burnin and reduce the length of the chain
burnin<-100
nout<-dim(result$rho)[1]
lim<-burnin:nout
#estimated CP consensus
cpcons<-CPconsensus(result$rho[lim,],labs)
names(cpcons)<-labs
rbind(latrank,cpcons)

#heat plot of rho (ordered according to true consensus)
matcons<-heatMat(result$rho[lim,],1,latrank)
heatPlot(matcons,latrank)

# Figure 10 of the paper
# Marginal for each item of being ranked among top-h
toph<-3
marg<-c()
for (i in 1:n){
  marg[i]<-mean(result$rho[,i]<toph+1)
}

# Marginal for each item for each assessor of being ranked among top-h
margUser<-matrix(NA,ncol=n,nrow=N)
for(j in 1:N){
  for (i in 1:n){
    margUser[j,i]<-mean(result$Raug[lim,j,i]<toph+1)
  }
}

margUser2<-rbind(marg,rep(NA,n),margUser)
margUser2<-margUser2[,order(margUser2[1,])]
par(mar=c(5.1,2.1,4.1,14.1))
image(margUser2, axes=F,main="Probability of being ranked among top-3",
      col=tim.colors(10000))
mtext(text=labs[order(marg)], side=2, line=0.3, at=seq(0,1,1/14), las=1, cex=1)
ini<-1/(N+2)
atSeq<-seq(0,1,ini)
mtext(text=c(expression(rho),"",paste("User",1:N)), side=1, 
      line=0.3, at=seq(0,1,1/(N+1)), las=2, cex=0.5)
par(mar=c(5.1,1.1,4.1,4.1))
image.plot(margUser2, col=tim.colors(10000),legend.only=TRUE, horizontal = F)

