rm(list=ls())
library(Rcpp)
library(RcppArmadillo)
library(fields)
library(gtools)
sourceCpp("MCMCfunction.cpp")
sourceCpp("MCMCsampler.cpp")

#Function that computes the CP consensus
CPconsensus<-function(mcmcOutput,labs){
  
  consensus <- vector(length=n)
  if(is.character(labs)){ items<-labs}else{
    paste("0",1:n,sep="")
  }
  names(consensus)<-items
  nmceff<-dim(mcmcOutput)[1]
  lim=(nmceff/10):nmceff
  nout <- length(lim)
  rem.items <- items
  dimnames(mcmcOutput)[[2]]<-items
  
  for(i in 1:(n-1)){
    chi<-colSums(mcmcOutput[lim,rem.items]<=i)/nout
    ind.item <- which.max(chi)
    which.item <- rem.items[ind.item]
    consensus[which.item] <- i
    rem.items <- rem.items[-ind.item]
  }
  consensus[rem.items] <- n
  
  return(consensus)
} 
#Function that computes matrix of posterior marginal probabilities for each item
heatMat<-function(mcmcOutput,burnin,t_rank){
  n<-dim(mcmcOutput)[2]
  lim1<-burnin:dim(mcmcOutput)[1]
  f<-matrix(nrow=n, ncol=n)
  colnames(mcmcOutput)<-NULL
  for(j in 1:n){
    ind <- which(t_rank==j)
    f[j, ] <- sapply(1:n, function(x) sum(mcmcOutput[lim1,ind] == x))
  }
  f <- f/length(lim1)
  return(f)
}
#Function that created the heatplot
heatPlot<-function(mat,t_rank){
  
  if(is.character(names(t_rank))){
    items<-names(sort(t_rank))}else{
      items<-paste("0",order(t_rank),sep="")
    }
  
  par(mfrow=c(1,1))
  par(mar=c(2,2,2,6))
  image(mat,col=tim.colors(64*10),zlim=range(mat),axes=F,cex.lab=2)
  
  mtext(text=items, side=1, line=0.3, 
        at=seq(0,1,1/(n-1)), las=1, cex=1)
  mtext(text=c(1:n), side=2, line=0.3, 
        at=seq(0,1,1/(n-1)), las=2, cex=1)
  par(mar=c(2,2,2,2))
  image.plot(mat, legend.only=T,horizontal = F)
}
#Function to sample from mixture of Mallows
samplerClusters<-function(latranks,alfas,N,psi,C,metric){
  n<-length(latranks[1,])
  R<-matrix(ncol=n,nrow=N)
  tau<-rdirichlet(1,rep(psi,C))
  zeta<-which(apply(t(rmultinom(N,1,tau)),1,"==",1),arr.ind = T)[,1]
  
  for(c in 1:C){
    chi<-which(zeta==c)
    Nc<-length(chi)
    pars <- list("nmc"=N*10^3+1, "thin"=10^3, "burnin"=10^5, "L"=n/2-1, 
                 "dist"=metric, "K"=Nc)
    R[chi,]<-MCMCSampler(latranks[c,],alfas[c],pars)$R 
  }
  
  return(list("R"=R,"tau"=tau,"zeta"=zeta))
}
#Function that returns a ranking, given an ordering 
Ord2Rank<-function(ordering){
  n<-length(ordering)
  ranking<-vector(length=n)
  names(ranking)<-paste("A",1:n,sep="")
  for(i in 1:n){
    ranking[ordering[i]]<-i
  }
  return(ranking)
}
#Function to sample partial rankings: 
#from Nxn matrix of full ranks, it returns a Nxn matrix with -1 if R_ij>=topk
R2partial<-function(R,topk){
  Rpart<-R
  Rpart[-c(which(is.element(R,1:topk),arr.ind=T))]<--1
  colnames(Rpart)<-paste("A",1:n,sep="")
  items<-c()
  for(j in 1:N) items <- c(items, names(which(Rpart[j,]!=-1)))
  items <- unique(items)
  n <- length(items)
  Rpart<-Rpart[,names(Ord2Rank(items))]
  return(Rpart)
}

###  Partial ranks (with clusters) ###  

#A# Sample N partial rankings from the Mallows model
#   with a given distance (metric) 
#   Modification of Algorithm 6 of the paper

# choice of the model parameters
N <- 50
n <- 10
K <- 2
alfas <- c(4,6)
latranks <- matrix(ncol=n,nrow=K)
for(c in 1:K) latranks[c,] <- sample(n,n)
labs <- paste("A",1:n,sep="")
colnames(latranks) <- labs
metric <- "footrule"
psi <- 10
# sample full ranks first
s1 <- samplerClusters(latranks,alfas,N,psi,K,metric)
R <- s1$R
z <- s1$zeta
tau <- s1$tau
# Choose topk (top-k rakings)
topk<-4
R<-R2partial(R,topk)

#B# Estimate the model parameters via MCMC 
#   Algorithm 3 of the paper

# choose MCMC parameters
nmc <- 1e5
thin <- 1e1
alphaJump <- 1e1
lambda <- 0.1
L <- n/5
sdAlpha <- 0.4

# partition function
if(metric=="Kendall"){# closed form available
  fitvec <- 0
  seq2 <- 0
}else if(metric=="footrule"){
  if(n <=50){# closed form available
    load("CdFootrule.Rdata")
    fitvec <- 0
  }else{# for larger n, run the IS
    seq2 <- 0
    sourceCpp("./importanceSampler.cpp")
    propDist   <- metric # metric in the Mallows model used in the pseudo-likelihood IS proposal
    targetDist <- metric # metric in the Mallows model used for computing Z
    KK <- 1e6           # number of IS iterations
    alpha     <- seq(.01,10,length.out=50) # values of alpha for which the IS approximation has to be computed
    ## parameters for the IS run:
    ## Z (defaults to a vector of 0s) is a pre-computed vector of partition function values on the same alpha grid
    ## and K is the corresponding number of IS iterations used to compute it
    pars <- list("nmc"=KK, "n"=n, "propDist"=propDist, "targetDist"=targetDist, "K"=0, "Z"=rep(0,length(alpha)))
    y <- importanceSampler(alpha, pars)    
    x <- alpha
    # approximate Z via a polynomial function of degree 10
    fitvec <- lm(y ~ x + I(x^2) + I(x^3) + I(x^4)+ I(x^5) +  I(x^6) + I(x^7) + I(x^8) + I(x^9) + I(x^10))$coefficients
  }
}else if (metric=="spearman"){# closed form available
  if(n <=14){
    load("CdSpearman.Rdata")
    fitvec <- 0
  }else{# for larger n, run the IS
    seq2 <- 0
    sourceCpp("./importanceSampler.cpp")
    propDist   <- metric
    targetDist <- metric
    KK <- 1e6
    alpha     <- seq(.01,20,length.out=50)
    pars <- list("nmc"=KK, "n"=n, "propDist"=propDist, "targetDist"=targetDist, "K"=0, "Z"=rep(0,length(alpha)))
    y <- importanceSampler(alpha, pars)    
    x <- alpha
    fitvec <- lm(y ~ x + I(x^2) + I(x^3) + I(x^4)+ I(x^5) +  I(x^6) + I(x^7) + I(x^8) + I(x^9) + I(x^10))$coefficients
  }
}


# run the MCMC
pars <- list("nmc"=nmc, "thin"=thin, "K"=K, "alphaJump"=alphaJump, 
             "lambda"=lambda, "L"=L, "sdAlpha"=sdAlpha, 
             "dist"=metric, "fit"=fitvec,"aug"=TRUE, 
             "Rmiss"=rep((topk+1),N),"alpha0"=rep(1,K), "rho0"=sample(n),
             "Cd"=seq2)
result <- MCMCclustering(R, pars)


#Acceptance
result$accRho
result$accAlpha
#Trace of Alpha
matplot(result$alpha,type = "l",lty=1)

#choose burnin and reduce the length of the chain
burnin<-200
nout<-dim(result$rho)[1]
lim<-burnin:nout
#estimated CP consensus
cpcons<-matrix(ncol=n,nrow=K)
for(k in 1:K){
  cpcons[k,]<-CPconsensus(result$rho[lim,,k],labs)}
colnames(cpcons)<-labs

cpcons
latranks

#heat plots of rho (ordered according to CP consensus of that cluster)
for(k in 1:K){
  matcons<-heatMat(result$rho[lim,,k],1,cpcons[k,])
  heatPlot(matcons,cpcons[k,])
}

#Posterior Assignment
nout<-length(lim)
predprobs <- matrix(0, N, K)
for(j in 1:N)predprobs[j, ] <- sapply(1:K, function(x) 
  sum(result$zeta[lim,j] == x))/nout
ZETAfinal <- numeric(N)
for(j in 1:N)ZETAfinal[j] <- which.max(predprobs[j,])
image(predprobs[sort.int(ZETAfinal,index.return=TRUE)$ix,], 
      col=tim.colors(64*10),axes=F, ylab="clusters", main="",xlab="assessors",cex.lab=1.2)
mtext(text=sort.int(ZETAfinal,index.return=TRUE)$ix, side=1, 
      line=0.3, at=seq(0,1,1/(N-1)), 
      las=1, cex=0.5)
mtext(text=1:K, side=2, line=0.3, at=seq(0,1,1/(K-1)), las=2)
par(mar=c(4,4,3,10))
image.plot(predprobs[sort.int(ZETAfinal,index.return=TRUE)$ix,], legend.only=T)

#Number of mis-assigned users (check labels!)
which(ZETAfinal!=z)



