#include <algorithm>
#include <RcppArmadillo.h>
#include <RcppArmadilloExtensions/sample.h>
// [[Rcpp::depends(RcppArmadillo)]]

using namespace Rcpp;
using namespace arma;

double dist(const double& R, const double& rho, const std::string& metric);
double dist(const vec& R, const vec& rho, const std::string& metric);
double dist(const mat& R, const mat& rho, const std::string& metric);
vec vecdist(const vec& R, const vec& rho, const std::string& metric);

// [[Rcpp::export]]
vec importanceSampler(vec alphaVec, List pars) {
  RNGScope scope;
  // Read in all the parameters //
  //****************************//
  int n = as<int>(pars["n"]);
  std::string metric1 = as<std::string>(pars["propDist"]);
  std::string metric2 = as<std::string>(pars["targetDist"]);
  // The dispersion parameter alpha comes as a vector value
  int nAlpha = alphaVec.n_elem;
  // The number of IS iterations is an integer
  int nmc = as<int>(pars["nmc"]);
  // A pre-computed vector of values of the normalizing constant (unscaled) - defaults to 0
  vec Zold = as<vec>(pars["Z"]);
  // The number of IS iterations needed to obtain Zold - defaults to 0
  int Kold = as<int>(pars["K"]);
  // (assumption: this refers to the same alpha vector as is being passed)
  
  double Z, q, alpha;
  vec myind, ranks, prob, support, cpd, rho=linspace<vec>(1,n,n), u;
  int jj, kMax, nmcEff=nmc-Kold;
  uvec inds;
  vec logZ = zeros(nAlpha); 
  
  for(int t=0; t<nAlpha; t++){
    alpha = alphaVec(t);
    Z = Zold(t);
    for(int i=0; i<nmcEff; i++){//compute the Mallows pseudo-likelihood proposal
      support = linspace<vec>(1,n,n);
      ranks = arma::zeros(n);
      q = 1; // probability of realized ranks
      u = randu(n,1);
      // Loop over possible values given to item j --> in RANDOM ORDER!!
      myind = shuffle(linspace(0, n-1, n));
      for(int j=0; j<n; j++){
        jj = myind(j);
        prob = zeros(n);
        inds = find(support != 0);
        kMax = inds.n_elem;
        prob(inds) = exp(-alpha/n*vecdist(inds+arma::ones(kMax),rho(jj)*arma::ones(kMax), metric1));
        prob = prob/accu(prob);
        cpd = cumsum(prob);
        // Draw a random sample
        inds = find(cpd > u(jj));
        ranks(jj) = inds(0)+1;
        q = q*prob(inds(0));
        support(ranks(jj)-1) = 0; // Remove the realized rank from support
      }
      Z += exp(-alpha/n*dist(ranks, rho, metric2))/q;
    }
    logZ(t)=log(Z/nmc);
  }
  return logZ;
}

double dist(const double& R, const double& rho, const std::string& metric){
  double totDist=0;
  if(metric == "footrule"){
    totDist = abs(R-rho);
  } else if(metric == "spearman") {
    totDist = pow(R - rho, 2.0);
  }
  return totDist;
}

double dist(const mat& R, const mat& rho, const std::string& metric){
  double totDist=0;
  int N = R.n_rows;
  int n = R.n_cols;
  if(metric == "footrule"){
    totDist = norm(R - rho, 1);
  }else if(metric == "kendall"){
    for(int i=0; i<N; i++){
      for(int u=0; u<n; u++){
        for(int t=0; t<u; t++){
          if((R(i,t) > R(i,u)) & (rho(i,t) < rho(i,u) ) ||
             ((R(i,t) < R(i,u) ) & (rho(i,t) > rho(i,u)))) totDist +=1;
        }
      }
    }
  } else if(metric == "spearman") {
    totDist = pow(norm(R - rho, 2),2.0);
  }
  return totDist;
}

double dist(const vec& R, const vec& rho, const std::string& metric){
  int n = rho.n_elem;
  double totDist=0;
  if(metric == "footrule"){
    totDist = norm(R - rho, 1);
  } else if(metric == "kendall"){
    for(int u=0; u<n; u++){
      for(int t=0; t<u; t++){
        if((R(t) > R(u)) & (rho(t) < rho(u) ) || ((R(t) < R(u) ) & (rho(t) > rho(u)))) totDist +=1;
      }
    }
  } else if(metric == "spearman") {
    totDist += pow(norm(R - rho, 2),2.0);
  }
  return totDist;
}

vec vecdist(const vec& R, const vec& rho, const std::string& metric){
  int n = rho.n_elem;
  vec totDist=zeros(n);
  if(metric == "footrule"){
    totDist = abs(R - rho);
  } else if(metric == "spearman") {
    totDist = pow(R - rho, 2.0);
  }
  return totDist;
}
